/**
 * @file <argos3/plugins/simulator/physics_engines/dynamics2d/dynamics2d_differentialsteering_control.cpp>
 *
 * @author Carlo Pinciroli - <ilpincy@gmail.com>
 */

#include "pointmass3d_mybot_model.h"
#include <argos3/core/utility/math/cylinder.h>

namespace argos {

   static const Real MYBOT_RADIUS                   = 0.085036758f;
   static const Real MYBOT_INTERWHEEL_DISTANCE      = 0.14f;
   static const Real MYBOT_HEIGHT                   = 0.146899733f;

   enum MYBOT_WHEELS {
      MYBOT_LEFT_WHEEL = 0,
      MYBOT_RIGHT_WHEEL = 1
   };

   /****************************************/
   /****************************************/

   CPointMass3DMyBotModel::CPointMass3DMyBotModel(CPointMass3DEngine& c_engine,
                                                      CMyBotEntity& c_mybot) :
      CPointMass3DModel(c_engine, c_mybot.GetEmbodiedEntity()),
      m_cWheeledEntity(c_mybot.GetWheeledEntity()),
      m_fCurrentWheelVelocity(m_cWheeledEntity.GetWheelVelocities()) {
      /* Register the origin anchor update method */
      RegisterAnchorMethod(GetEmbodiedEntity().GetOriginAnchor(),
                           &CPointMass3DMyBotModel::UpdateOriginAnchor);
      /* Get initial rotation */
      CRadians cTmp1, cTmp2;
      GetEmbodiedEntity().GetOriginAnchor().Orientation.ToEulerAngles(m_cYaw, cTmp1, cTmp2);
   }

   /****************************************/
   /****************************************/

   void CPointMass3DMyBotModel::Reset() {
      CPointMass3DModel::Reset();
      CRadians cTmp1, cTmp2;
      GetEmbodiedEntity().GetOriginAnchor().Orientation.ToEulerAngles(m_cYaw, cTmp1, cTmp2);
      m_fAngularVelocity = 0.0;
   }

   /****************************************/
   /****************************************/


   void CPointMass3DMyBotModel::UpdateFromEntityStatus() {
      m_cVelocity.Set((m_fCurrentWheelVelocity[MYBOT_RIGHT_WHEEL] + m_fCurrentWheelVelocity[MYBOT_LEFT_WHEEL])*0.5, 0.0, 0.0);
      m_cVelocity.RotateZ(m_cYaw);
      m_fAngularVelocity = (m_fCurrentWheelVelocity[MYBOT_RIGHT_WHEEL] - m_fCurrentWheelVelocity[MYBOT_LEFT_WHEEL]) / MYBOT_INTERWHEEL_DISTANCE;
   }

   /****************************************/
   /****************************************/

   void CPointMass3DMyBotModel::Step() {
      m_cPosition += m_cVelocity * m_cPM3DEngine.GetPhysicsClockTick();
      m_cYaw += CRadians(m_fAngularVelocity * m_cPM3DEngine.GetPhysicsClockTick());
   }

   /****************************************/
   /****************************************/

   void CPointMass3DMyBotModel::CalculateBoundingBox() {
      GetBoundingBox().MinCorner.Set(
         GetEmbodiedEntity().GetOriginAnchor().Position.GetX() - MYBOT_RADIUS,
         GetEmbodiedEntity().GetOriginAnchor().Position.GetY() - MYBOT_RADIUS,
         GetEmbodiedEntity().GetOriginAnchor().Position.GetZ());
      GetBoundingBox().MaxCorner.Set(
         GetEmbodiedEntity().GetOriginAnchor().Position.GetX() + MYBOT_RADIUS,
         GetEmbodiedEntity().GetOriginAnchor().Position.GetY() + MYBOT_RADIUS,
         GetEmbodiedEntity().GetOriginAnchor().Position.GetZ() + MYBOT_HEIGHT);
   }

   /****************************************/
   /****************************************/

   bool CPointMass3DMyBotModel::CheckIntersectionWithRay(Real& f_t_on_ray,
                                                           const CRay3& c_ray) const {
      CCylinder m_cShape(MYBOT_RADIUS,
                         MYBOT_HEIGHT,
                         m_cPosition,
                         CVector3::Z);
      bool bIntersects = m_cShape.Intersects(f_t_on_ray, c_ray);
      return bIntersects;
   }

   /****************************************/
   /****************************************/

   void CPointMass3DMyBotModel::UpdateOriginAnchor(SAnchor& s_anchor) {
      s_anchor.Position = m_cPosition;
      s_anchor.Orientation = CQuaternion(m_cYaw, CVector3::Z);
   }

   /****************************************/
   /****************************************/
   
   REGISTER_STANDARD_POINTMASS3D_OPERATIONS_ON_ENTITY(CMyBotEntity, CPointMass3DMyBotModel);

   /****************************************/
   /****************************************/

}
